# forked from http://pypi.python.org/pypi/AddOns
import sys
from weakref import ref
from mext.classhacks import classy, on_class_ready, is_in_class_declaration

__all__ = ['AddOn', 'ClassAddOn', 'Registry', 'addons_for']

_addons = {}

class AddonsIncapable(TypeError):
    pass

def addons_for(ob):
    """Get the dictionary that should contain add-ons for `ob`"""
    try:
        r = ob.__dict__
        r.setdefault # classes' __dict__ is not changeable
        return r
    except (AttributeError, TypeError):
        try:
            return _addons[ref(ob)]
        except TypeError:
            raise AddonsIncapable("%s has no __dict__ and no __weakref__ slot" % ob)
        except KeyError:
            return _addons.setdefault(ref(ob, _addons.__delitem__), {})

#_addons = dict((k,v) for (k,v) in _addons if k() is not None)


class AddOn(classy):
    """Attach extra state to (almost) any object"""
    __slots__ = ()

    @classmethod
    def __class_call__(cls, ob, *data):
        a = addons_for(ob)
        addon_key = cls.addon_key(*data)
        try:
            return a[addon_key]
        except KeyError:
            return a.setdefault(addon_key, super(AddOn, cls).__class_call__(ob, *data))

    @classmethod
    def addon_key(cls, *args):
        if args: return (cls,)+args
        return cls

    @classmethod
    def exists_for(cls, ob, *key):
        """Does an aspect of this type for the given key exist?"""
        try:
            return cls.addon_key(*key) in addons_for(ob)
        except AddonsIncapable:
            return False

    @classmethod
    def delete_from(cls, ob, *key):
        """Ensure an aspect of this type for the given key does not exist"""
        addons_for(ob).pop(cls.addon_key(*key), None)

    def __init__(self, subject):
        pass




class ClassAddOn(AddOn):
    """Attachment/annotation for classes and types"""
    __slots__ = ()

    @classmethod
    def __class_call__(cls, ob, *data):
        addon_key = cls.addon_key(*data)
        try:
            return ob.__dict__[addon_key]
        except KeyError:
            pass
        a = addons_for(ob)
        try:
            return a[addon_key]
        except KeyError:
            return a.setdefault(
                addon_key,
                super(ClassAddOn, cls).__class_call__(ob, *data)
            )

    @classmethod
    def for_enclosing_class(cls, *args, **kw):
        depth = kw.pop('depth', 1) + 1
        if kw:
            raise TypeError("Unexpected keyword arguments", kw)
        elif not is_in_class_declaration(depth):
            raise SyntaxError("Not in class declaration")
        a = sys._getframe(depth).f_locals
        addon_key = cls.addon_key(*args)
        try:
            return a[addon_key]
        except KeyError:
            r = a[addon_key] = type.__call__(cls, None, *args)
            # we use a lambda here so that if we are a registry, Python 2.5
            # won't consider our method equal to some other registry's method
            on_class_ready(lambda cls: r.created_for(cls), depth=depth)
            return r

    @classmethod
    def exists_for(cls, ob, *key):
        """Does an aspect of this type for the given key exist?"""
        addon_key = cls.addon_key(*key)
        return addon_key in ob.__dict__ or addon_key in addons_for(ob)

    @classmethod
    def delete_from(cls, ob, *key):
        """Class AddOns are not deletable!"""
        raise TypeError("ClassAddOns cannot be deleted")

    def created_for(self, cls):
        """Override to access the decorated class, as soon as it's known"""

    def __init__(self, subject):
        """Ensure ``created_for()`` is called, if class already exists"""
        if subject is not None:
            self.created_for(subject)





class Registry(ClassAddOn, dict):
    """ClassAddOn that's a dictionary with mro-based inheritance"""
    __slots__ = ()

    def __new__(cls, subject):
        if cls is Registry:
            raise TypeError("You must subclass Registry to use it")
        return super(Registry, cls).__new__(cls)

    def __init__(self, subject):
        dict.__init__(self)
        super(Registry, self).__init__(subject)

    def created_for(self, cls):
        """Inherit the contents of base classes"""
        #@@ if the baseclass registry changes, those changes aren't propagated
        self.defined_in_class = dict(self)
        try:
            mro = cls.__mro__[::-1]
        except AttributeError:
            mro = type(cls.__name__, (cls,object), {}).__mro__[1:][::-1]
        map(self.update, map(type(self), mro[:-1]))
        self.update(self.defined_in_class)

    def set(self, key, value):
        if self.setdefault(key, value) != value:
            raise ValueError("%s[%r] already contains %r; can't set to %r"
                % (self.__class__.__name__, key, self[key], value)
            )




