import sys
from weakref import ref
from types import MethodType, FunctionType, ClassType

__all__ = []


#
# get_result(f, *args, **kw)
#


def get_result(__f, *args, **kw):
    """
        Return a result of calling `__f(*args, **kw)`
        as a `Result` or `Failure`. This never throws exceptions.
    """
    try:
        return Result(__f(*args, **kw))
    except:
        return Failure()

# Result and Failure are callables -- call to get the
# pending result / exception.
# Both have .exc_info attr.

def Result(value):
    r = lambda: value
    r.exc_info = None
    return r

def Failure(exc=None):
    if exc is None:
        exc = sys.exc_info()
    if isinstance(exc, tuple):
        def reraise():
            raise exc[0], exc[1], exc[2]
        reraise.exc_info = exc
    else:
        assert isinstance(exc, Exception)
        def reraise():
            raise exc
        reraise.exc_info = type(exc), exc, None
    return reraise


#
# `classy` implementation
#


def super_next(cls, attr):
    for c in cls.__mro__:
        if attr in c.__dict__:
            yield getattr(c, attr).im_func

class classy_class(type):
    """Metaclass that delegates selected operations back to the class"""

    def __new__(meta, name, bases, cdict):
        cls = super(classy_class, meta).__new__(meta, name, bases, cdict)
        supr = super_next(cls, '__class_new__').next
        return supr()(meta, name, bases, cdict, supr)

    def __init__(cls, name, bases, cdict):
        supr = super_next(cls, '__class_init__').next
        return supr()(cls, name, bases, cdict, supr)

    def __call__(cls, *args, **kw):
        return cls.__class_call__.im_func(cls, *args, **kw)


class classy(object):
    """Base class for classes that want to be their own metaclass"""
    __metaclass__ = classy_class
    __slots__ = ()

    def __class_new__(meta, name, bases, cdict, supr):
        return type.__new__(meta, name, bases, cdict)

    def __class_init__(cls, name, bases, cdict, supr):
        return type.__init__(cls, name, bases, cdict)

    def __class_call__(cls, *args, **kw):
        return type.__call__(cls, *args, **kw)
    __class_call__ = classmethod(__class_call__)




#
# Hacks to alter classes that are not yet fully declared
#

def is_in_class_declaration(depth):
    try:
        return '__module__' in sys._getframe(depth+1).f_locals
    except ValueError:
        return False



def on_class_ready(cb, depth=1):
    frame = sys._getframe(depth+1)
    if '__module__' not in frame.f_locals:
        raise SyntaxError("Not in class declaration")
    if '__metaclass__' in frame.f_globals:
        raise SyntaxError("Module-wide metaclasses are not supported")

    meta_prev = frame.f_locals.get('__metaclass__')
    callbacks = getattr(meta_prev, '_on_class_ready_callbacks', None)
    if callbacks is not None:
        callbacks.append(cb)
        return

    def meta_cb(name, bases, cdict):
        cdict.pop('__metaclass__', None)
        meta = metaclass_for_bases(bases, meta_prev)
        r = meta(name, bases, cdict)
        for cb in callbacks:
            cb(r)
        return r

    callbacks = meta_cb._on_class_ready_callbacks = [cb]
    frame.f_locals['__metaclass__'] = meta_cb


def metaclass_for_bases(bases, explicit_mc=None):
    meta = map(type, bases)
    if explicit_mc is not None:
        meta.append(explicit_mc)
    if ClassType in meta:
        raise TypeError("Classic classes are not supported: %s" % [b for b in bases if isinstance(b, ClassType)])
    r = None
    for m in meta:
        if r is None or issubclass(m, r):
            r = m
        elif not issubclass(r, m):
            raise TypeError("Incompatible metatypes", bases)
    return r or ClassType


# def init_args(*argnames): #@@ less magic
#     """
#         Adds constructor to this class to convert positional arguments to keyword args
#         based on `argnames` argument. Passes the resulting kw to superclass __init__
#     """
#     @on_class_ready
#     def set_init(cls):
#         def init(self, *args, **kw):
#             for name, val in zip(argnames, args):
#                 if name in kw:
#                     raise ValueError('Duplicate argument: %s' % name)
#                 kw[name] = val
#             super(cls, self).__init__(**kw)
#         cls.__init__ = init








def get_name(ob, default=TypeError):
    if isinstance(ob, (type, ClassType)):
        return ob.__name__
    elif isinstance(ob, FunctionType):
        return ob.func_name
    elif isinstance(ob, MethodType):
        return get_name(ob.im_func, default)
    elif isinstance(ob, (classmethod, staticmethod)):
        return get_name(ob.__get__(None, object), default)
    elif isinstance(ob, property):
        return get_name(ob.fget, default)
    elif default is not TypeError:
        return default
    else:
        raise TypeError("Can't get name of %r" % ob)









#
# Method injection helpers
#

def inject(parent, setname=None, save=False):
    def deco(ob):
        if setname is None:
            name = get_name(ob)
        else:
            name = setname
        if save:
            if isinstance(save, str):
                prev_name = save
            else:
                prev_name = '_%s_%s' % (parent.__name__, name)
            setattr(parent, prev_name, getattr(parent, name))
        setattr(parent, name, ob)
        return ob
    return deco



def mix_into(cls, mixin=None):
    if mixin is not None:
        cls.__bases__ = (mixin,) + cls.__bases__
    else:
        on_class_ready(lambda mixin: mix_into(cls, mixin))























#
# Utility property classes
#


def bind(rule, ob):
    if hasattr(rule, '__get__'):
        return rule.__get__(ob, type(ob))
    return rule


def _getattr_path(ob, path):
    for name in path:
        ob = getattr(ob, name)
    return ob

def _setattr_path(ob, path, value):
    ob = _getattr_path(ob, path[:-1])
    setattr(ob, path[-1], value)

def _norm_attrpath(path):
    if isinstance(path, str):
        return tuple(path.split('.'))
    elif isinstance(path, (list, tuple)):
        return tuple(path)
    else:
        raise TypeError(path)

class alias(object):
    __slots__ = 'path'
    def __init__(self, path):
        self.path = _norm_attrpath(path)

    def __repr__(self):
        return '<%s(%r)>' % (self.__class__.__name__, '.'.join(self.path))

    def __get__(self, ob, type):
        if ob is None:
            return self
        return _getattr_path(ob, self.path)

    def __set__(self, ob, val):
        _setattr_path(ob, self.path, val)


class filter_alias(object):
    __slots__ = 'path', 'decode', 'encode'
    def __init__(self, path, decode, encode):
        self.path = _norm_attrpath(path)
        self.decode = decode
        self.encode = encode

    def __get__(self, ob, typ):
        if ob is None:
            return self
        val = _getattr_path(ob, self.path)
        decode = bind(self.decode, ob)
        return decode(val)

    def __set__(self, ob, val):
        encode = bind(self.encode, ob)
        val = encode(val)
        _setattr_path(ob, self.path, value)


