import sys
from types import ModuleType

__all__ = [
    'onimport', 'onimport_mod', 'onimport_cls',
    'autoonimport',
    'lazy_import', 'import_string',
]

modules = sys.modules
_mod_hooks = {}


def import_string(name):
    mname = name
    while True:
        try:
            r = __import__(mname)
            break
        except ImportError:
            if '.' not in mname:
                raise
            mname = mname.rsplit('.', 1)[0]

    for seg in name.split('.')[1:]:
        try:
            r = getattr(r, seg)
        except AttributeError:
            raise ImportError("%r has no %r attribute" % (r, seg))

    if isinstance(r, LazyModule):
        _load_mod(r)
    return r




def lazy_import(modname):
    if modname not in modules:
        mtype = type('LazyModule_%s'%modname, (LazyModule,), {})
        modules[modname] = mtype(modname)
        if '.' in modname:
            # ensure parent module/package is in sys.modules
            # and parent.modname=module, as soon as the parent is imported
            splitpos = modname.rindex('.')
            pname, subname = modname.rsplit('.', 1)
            onimport_mod(pname, lambda m: setattr(m, subname, modules[modname]))

    return modules[modname]


class LazyModule(ModuleType):
    __slots__ = ()
    def __init__(self, name):
        ModuleType.__setattr__(self, '__name__', name)
        #super(LazyModule, self).__init__(name)

    def __getattribute__(self, attr):
        _load_mod(self)
        return ModuleType.__getattribute__(self, attr)

    def __setattr__(self, attr, value):
        _load_mod(self)
        return ModuleType.__setattr__(self, attr, value)


def _load_mod(module):
    mtype = type(module)
    mtype.__getattribute__ = ModuleType.__getattribute__
    mtype.__setattr__ = ModuleType.__setattr__
    assert module.__dict__.keys() == ['__name__']

    reload(module)

    try:
        for hook in get_mod_hooks(module.__name__):
            hook(module)
    finally:
        # Ensure hooks are not called again, even if they fail
        _mod_hooks[module.__name__] = None


def get_mod_hooks(name):
    hooks = _mod_hooks.setdefault(name, [])
    if hooks is None:
        raise RuntimeError("Module already imported", name)
    return hooks


def _set_mod_hook(name, hook):
    if name in modules and _mod_hooks.get(name) is None:
        # Module is already imported/loaded, just call the hook
        module = modules[name]
        hook(module)
        return module

    get_mod_hooks(name).append(hook)
    return lazy_import(name)


















def onimport_mod(name, hook):
    if '.' in name:
        pname, sname = name.rsplit('.', 1)
        onimport_mod(pname, lambda m: _set_mod_hook(name, hook))
    else:
        _set_mod_hook(name, hook)





def autodecorate(func):
    def autodecorate_wrapped(*args):
        if len(args) > 1:
            return func(*args)
        else:
            arg0 = args[0]
            return lambda *args: func(arg0, *args)
    return autodecorate_wrapped

@autodecorate
def onimport_cls(path, cb, *args):
    path, name = path.rsplit('.', 1)
    def import_cb(mod):
        cls = getattr(mod, name)
        cb(cls, *args)
    onimport_mod(path, import_cb)
    return cb

def onimport(path, *extra):
    def decorate(func):
        if extra:
            def cb(mod):
                onimport(*extra)(func)
        else:
            cb = func
        onimport_mod(path, cb)
        return func
    return decorate


def autoonimport(func):
    def autoonimport_wrapped(cls, *args, **kw):
        if isinstance(cls, str):
            onimport_cls(cls, lambda cls: func(cls, *args, **kw))
        else:
            return func(cls, *args, **kw)
    return autoonimport_wrapped


## class LazyLib(object):
##     def __init__(self, parts=()):
##         self.parts = parts
##     def __getattr__(self, item):
##         return LazyLib(self.parts+(item,))
##     def __call__(self):
##         return lazy_import('.'.join(self.parts))
##
## lazylib = LazyLib()
##
